<?php
/**
 * Created by PhpStorm.
 * User: felipesantos
 * Date: 01/09/2015
 * Time: 15:18
 */

namespace Service;


use Doctrine\ORM\EntityManager;
use Service\Hcm\CalculoService;
use Service\Hcm\ContabilService;
use Service\Hcm\EmpresaService;
use Service\Hcm\EventoService;
use Service\Hcm\FinanceiroService;
use Service\Hcm\GeralService;
use Service\Hcm\PessoaService;
use Service\Ems\GestorService;
use Util\Environment;
use Util\DateUtil;
use Util\NovoCodigo;


class ApiHcmServices {

    const TPESTRUT_TOMADOR = '7';

    const CDOPERADOR = '2';

	private $environment;
    private $pessoaService;
    private $empresaService;
    private $calculoService;
    private $contabilService;
    private $financeiroService;
    private $eventoService;

    public function __construct(EntityManager $entityManager, Environment $environment, GeralService $geralService, PessoaService $pessoaService, EmpresaService $empresaService, CalculoService $calculoService,
                                ContabilService $contabilService, FinanceiroService $financeiroService, EventoService $eventoService){
        $this->entityManager = $entityManager;
        $this->environment = $environment;
        $this->geralService = $geralService;
        $this->pessoaService = $pessoaService;
        $this->empresaService = $empresaService;
        $this->calculoService = $calculoService;
        $this->contabilService = $contabilService;
        $this->financeiroService = $financeiroService;
        $this->eventoService = $eventoService;
    }

    public function findItensPagamentoByCnpjAndDataAndMovimento($cnpj, $data, $movimento, $banco, $cdTpCtPagar, $cnpjFornecedor, $dataCompetencia){
        try {
            $result = array();
            $obsPagamento = '';

            $nrorg = $this->environment->getNrOrg();
            $dtocorrencia = !is_null($data) ? DateUtil::getDataDeString($data)->format('d/m/Y') : null;
            $dtmescompetencia = !is_null($data) ? DateUtil::getDataDeString($data)->format('01/m/Y') : DateUtil::getDataDeString($dataCompetencia)->format('01/m/Y');
            $dtcompetencia = !is_null($dataCompetencia) ? DateUtil::getDataDeString($dataCompetencia)->format('01/m/Y') : null;

            $nrtipomoviment = $this->calculoService->getNumeroTipoMovimento($movimento);
            if (is_null($nrtipomoviment))
                throw new \Exception("Tipo de pagamento informado não encontrado.");

            $nreventoLiq = $this->calculoService->getEventoLiquidoMovimento($movimento);

            // Verifica se existe associação financeira para o fornecedor ou o tipo de conta informados
            $associacaoFinanc = $this->financeiroService->getAssociacaoFinanceira($nrorg, $cnpjFornecedor, $cdTpCtPagar);
            
            $result = array();
            
            $paramsOrg = $this->geralService->getParametrosOrg($nrorg);
            $somenteFechado = is_array($paramsOrg) && $paramsOrg['IDAPROVALANCPAG'] == 'S' ? true : false;

            if (empty($associacaoFinanc)) {
                /* Retorna pagamentos por vínculo */
                $estrutura = $this->empresaService->getEstruturalegalByCnpjAndCompetenciaAndNrorg($cnpj, $dtcompetencia, $nrorg);
                if (empty($estrutura))
                    throw new \Exception("O CNPJ $cnpj informado não foi encontrado.");

                $vinculos = $this->pessoaService->getVinculosByEstruturaAndCompetenciaAndNrorg($estrutura['NRESTRUTURAM'], $dtcompetencia, $nrorg);
                if (empty($vinculos))
                    throw new \Exception("Não foram encontrados funcionários para o CNPJ $cnpj informado.");

                foreach ($vinculos as $vinculo) {
                    $arrayCentCust = array();
                    $arrayFiliais = array();

                    $pessoa = $this->pessoaService->getPessoaByNrpessoaAndCompetenciaAndNrorg($vinculo['NRPESSOA'], $dtcompetencia, $nrorg);

                    $arraycontas = array();
                    $contas = $this->pessoaService->getContaPessoaByNrpessoaAndNrorg($vinculo['NRPESSOA'], $nrorg, $banco);

                    if (!is_null($banco) and empty($contas)) {
                        continue;
                    }

                    foreach ($contas as $conta) {
                        $contapessoa['banco'] = $conta['CDBANCO'];
                        $contapessoa['banco_nome'] = $conta['NMBANCO'];
                        $contapessoa['agencia'] = $conta['CDAGENCIA'];
                        $contapessoa['conta'] = $conta['CDCONTCORR'];
                        $contapessoa['tipo'] = $conta['DSTPCONTCORR'];
                        $arraycontas[] = $contapessoa;
                        $contapessoa = null;
                    }

                    $itensCalculo = $this->calculoService->getItensByMovimentoAndEventoAndDtocorrenciaAndVinculoAndNrorg($nrtipomoviment, $nreventoLiq, $dtocorrencia, $vinculo['NRVINCULOM'], $nrorg, $dtcompetencia, $somenteFechado, 'S');
                    
                    foreach ($itensCalculo as $itemCalculo) {

                        if (!is_null($vinculo['CDEXPCENTCUST'])) {
                            if (!isset($arrayCentCust[$vinculo['CDEXPCENTCUST']])) {
                                $arrayCentCust[$vinculo['CDEXPCENTCUST']] = 0;
                            }
                            $arrayCentCust[$vinculo['CDEXPCENTCUST']] += $itemCalculo['VREVENTOFOLHA'];
                        }
                        if (!is_null($vinculo['CDEXPFILIAL'])) {
                            if (!isset($arrayFiliais[$vinculo['CDEXPFILIAL']])) {
                                $arrayFiliais[$vinculo['CDEXPFILIAL']] = array();
                            }
                            if (!is_null($vinculo['CDEXPCENTCUST'])) {
                                if (!isset($arrayFiliais[$vinculo['CDEXPFILIAL']][$vinculo['CDEXPCENTCUST']])) {
                                    $arrayFiliais[$vinculo['CDEXPFILIAL']][$vinculo['CDEXPCENTCUST']] = 0;
                                }
                                $arrayFiliais[$vinculo['CDEXPFILIAL']][$vinculo['CDEXPCENTCUST']] += $itemCalculo['VREVENTOFOLHA'];
                            }
                        }

                        $pagamento = array();

                        $pagamento['cpf'] = isset($pessoa['NRCPFPESSOA']) ? $pessoa['NRCPFPESSOA'] : 'erro_cpf';
                        $pagamento['nome'] = isset($pessoa['NMPESSOA']) ? $pessoa['NMPESSOA'] : 'erro_cpf';
                        $pagamento['data'] = $itemCalculo['DTOCORRENCIA'];
                        $pagamento['contas'] = $arraycontas;
                        $pagamento['valor'] = $itemCalculo['VREVENTOFOLHA'];
                        if (!empty($arrayCentCust)) {
                            foreach ($arrayCentCust as $key => $value) {
                                $arrayCentCust[$key] = (string)$value;
                            }
                            $pagamento['centrocustos'] = $arrayCentCust;
                        }
                        if (!empty($arrayFiliais)) {
                            foreach ($arrayFiliais as $cdfilial => $centroscusto) {
                                foreach ($centroscusto as $key => $value) {
                                    $arrayFiliais[$cdfilial][$key] = (string)$value;
                                }
                            }
                            $pagamento['filiais'] = $arrayFiliais;
                        }

                        $result[] = $pagamento;
                        $pagamento = null;
                    }

                    $arraycontas = null;
                }
            } else {
                /* Retorna pagamentos sumarizados para o fornecedor encontrado */
                $valorTotal = 0;
                $arrayCentCust = array();
                $arrayFiliais = array();

                $estrutura = $this->empresaService->getEstruturalegalByCnpjAndCompetenciaAndNrorg($cnpj, $dtcompetencia, $nrorg);
                if (empty($estrutura))
                    throw new \Exception("O CNPJ $cnpj informado não foi encontrado.");
                
                // Busca vínculos pelas estruturas informadas na associação ou por estrutura legal conforme o CNPJ informado na requisição
                if (is_null($associacaoFinanc['NRESTRUTLEGAL']) && is_null($associacaoFinanc['NRESTRUTGEREN']) && is_null($associacaoFinanc['NRESTRUTSIND'])) {
                    $vinculos = $this->pessoaService->getVinculosByEstruturaAndCompetenciaAndNrorg(
                        $estrutura['NRESTRUTURAM'], $dtcompetencia, $nrorg, $associacaoFinanc['NRTIPOVINCULOM'], $associacaoFinanc['IDVINCRESCINDIDOS']
                    );
                    if (empty($vinculos))
                        throw new \Exception("Não foram encontrados funcionários para o CNPJ $cnpj informado.");
                } elseif (!is_null($associacaoFinanc['NRESTRUTSIND']) && is_null($associacaoFinanc['NRESTRUTLEGAL'])) {
                    $vinculos = $this->pessoaService->getVinculosByEstruturasAndCompetenciaAndNrorg(
                        $estrutura['NRESTRUTURAM'], $associacaoFinanc['NRESTRUTGEREN'], $associacaoFinanc['NRESTRUTSIND'], $dtcompetencia, $nrorg, $associacaoFinanc['NRTIPOVINCULOM'], $associacaoFinanc['IDVINCRESCINDIDOS']
                    );
                    if (empty($vinculos))
                        throw new \Exception("Não foram encontrados funcionários para o CNPJ $cnpj informado.");
                } else {
                    $vinculos = $this->pessoaService->getVinculosByEstruturasAndCompetenciaAndNrorg(
                        $associacaoFinanc['NRESTRUTLEGAL'], $associacaoFinanc['NRESTRUTGEREN'], $associacaoFinanc['NRESTRUTSIND'], $dtcompetencia, $nrorg, $associacaoFinanc['NRTIPOVINCULOM'], $associacaoFinanc['IDVINCRESCINDIDOS']
                    );
                    if (empty($vinculos))
                        throw new \Exception("Não foram encontrados funcionários para as estruturas informadas na associação financeira do HCM.");
                }
                
                // Soma os itens dos vínculos
                foreach ($vinculos as $vinculo) {
                    // Verifica se o vínculo possui conta para o banco, caso informado
                    $contas = $this->pessoaService->getContaPessoaByNrpessoaAndNrorg($vinculo['NRPESSOA'], $nrorg, $banco);
                    if (!is_null($banco) and empty($contas)) {
                        continue;
                    }

                    // Busca os itens conforme a incidência da associação, caso não tenha incidência busca pelo evento de líquido do movimento
                    if (is_null($associacaoFinanc['NRTIPOINCIDE'])) {
                        $itensCalculo = $this->calculoService->getItensByMovimentoAndEventoAndDtocorrenciaAndVinculoAndNrorg(
                            $nrtipomoviment, $nreventoLiq, $dtocorrencia, $vinculo['NRVINCULOM'], $nrorg, $dtcompetencia, $somenteFechado, $associacaoFinanc['IDUTLDTOCORRENCIA'], $associacaoFinanc['NROCORRECAL']
                        );
                    } else {
                        $itensCalculo = $this->calculoService->getItensByMovimentoAndIncidenciaAndDtocorrenciaAndVinculoAndNrorg(
                            $dtcompetencia, $nrtipomoviment, $associacaoFinanc['NRTIPOINCIDE'], $dtocorrencia, $vinculo['NRVINCULOM'], $nrorg, $somenteFechado, $associacaoFinanc['IDUTLDTOCORRENCIA'], $associacaoFinanc['NROCORRECAL']
                        );
                    }
                    
                    foreach ($itensCalculo as $itemCalculo) {
                        if (!is_null($vinculo['CDEXPCENTCUST'])) {
                            if (!isset($arrayCentCust[$vinculo['CDEXPCENTCUST']])) {
                                $arrayCentCust[$vinculo['CDEXPCENTCUST']] = 0;
                            }
                            $arrayCentCust[$vinculo['CDEXPCENTCUST']] += $itemCalculo['VREVENTOFOLHA'];
                        }
                        if (!is_null($vinculo['CDEXPFILIAL'])) {
                            if (!isset($arrayFiliais[$vinculo['CDEXPFILIAL']])) {
                                $arrayFiliais[$vinculo['CDEXPFILIAL']] = array();
                            }
                            if (!is_null($vinculo['CDEXPCENTCUST'])) {
                                if (!isset($arrayFiliais[$vinculo['CDEXPFILIAL']][$vinculo['CDEXPCENTCUST']])) {
                                    $arrayFiliais[$vinculo['CDEXPFILIAL']][$vinculo['CDEXPCENTCUST']] = 0;
                                }
                                $arrayFiliais[$vinculo['CDEXPFILIAL']][$vinculo['CDEXPCENTCUST']] += $itemCalculo['VREVENTOFOLHA'];
                            }
                        }
                        $valorTotal += $itemCalculo['VREVENTOFOLHA'];
                    }
                    
                    if (count($itensCalculo) > 0) {
                        $obsPagamento .= "\n" . $vinculo['NRVINCULOM'] . " - " . $vinculo['NMVINCULOM'];
                    }
                }

                if ($valorTotal > 0) {
                    
                    $arraycontas = array();
                    $contas = $this->pessoaService->getContaCorrenteByNrparceiroAndNrorg($associacaoFinanc['NRPARCNEGOCIO'], $nrorg, $banco);

                    foreach ($contas as $conta) {
                        $contafornecedor['banco'] = $conta['CDBANCO'];
                        $contafornecedor['banco_nome'] = $conta['NMBANCO'];
                        $contafornecedor['agencia'] = $conta['CDAGENCIA'];
                        $contafornecedor['conta'] = $conta['CDCONTCORR'];
                        $contafornecedor['tipo'] = $conta['DSTPCONTCORR'];
                        $arraycontas[] = $contafornecedor;
                        $contafornecedor = null;
                    }

                    $pagamento = array();

                    $pagamento['cpf'] = !is_null($associacaoFinanc['NRINSCRICAOPARC']) ? $associacaoFinanc['NRINSCRICAOPARC'] : 'erro_cpf';
                    $pagamento['nome'] = isset($associacaoFinanc['NMPRINCIPALPARC']) ? $associacaoFinanc['NMPRINCIPALPARC'] : 'erro_cpf';
                    $pagamento['data'] = DateUtil::getDataDeString((is_null($dtocorrencia) ? $dtmescompetencia : $dtocorrencia), DateUtil::FORMATO_BRASILEIRO, true)->format('Y-m-d H:i:s');
                    $pagamento['contas'] = $arraycontas;
                    $pagamento['valor'] = (string)$valorTotal;
                    foreach ($arrayCentCust as $key => $value) {
                        $arrayCentCust[$key] = (string)$value;
                    }
                    if (!empty($arrayCentCust)) {
                        $pagamento['centrocustos'] = $arrayCentCust;
                    }
                    foreach ($arrayFiliais as $cdfilial => $centroscusto) {
                        foreach ($centroscusto as $key => $value) {
                            $arrayFiliais[$cdfilial][$key] = (string)$value;
                        }
                    }
                    if (!empty($arrayFiliais)) {
                        $pagamento['filiais'] = $arrayFiliais;
                    }
                    
                    // Se for movimento de férias ou rescisão retorna descrição para a conta a pagar
                    if ($movimento == '2' || $movimento == '4') {
                        $pagamento['observacao'] = "Competência: $dtcompetencia" . $obsPagamento;
                    }

                    $result[] = $pagamento;
                }
            }

            if (empty($result))
                throw new \Exception("Não foram encontrados pagamentos para os parâmetros informados.");

            return $result;
        }
        catch (\Exception $e) {
            throw $e;
        }
    }

    public function retornaLancamentosDeCalculo($dtmescompetenc, $nrtipomoviment, $nrocorrecal, $cdcnpjestrut, $cdfilial, $cdcentcust, $nrcpfpessoa, $nreventom, $dtocorrencia){
        try {
            $result = array();

            $nrorg = $this->environment->getNrOrg();
            $dtmescompetenc = !is_null($dtmescompetenc) ? DateUtil::getDataDeString($dtmescompetenc)->format('01/m/Y') : DateUtil::getDataAtual(true)->format('01/m/Y');
            $dtocorrencia = !is_null($dtocorrencia) ? DateUtil::getDataDeString($dtocorrencia)->format('d/m/Y') : null;
            
            /* Retorna lançamentos */
            $itens = $this->calculoService->getItensByEstruturasCalculoItemVinculoEvento(
                $nrorg, $dtmescompetenc, $nrtipomoviment, $nrocorrecal, $dtocorrencia, $nreventom, $nrcpfpessoa, $cdcnpjestrut, $cdfilial, $cdcentcust
            );

            if (empty($itens))
                throw new \Exception("Não foram encontrados pagamentos para os parâmetros informados.");

            return $itens;
        }
        catch (\Exception $e) {
            throw $e;
        }
    }

    public function findResumoContabilByCnpjAndMes($cnpj, $mes)
    {
        try {
            $result = array();

            $nrorg = $this->environment->getNrOrg();
            $dtocorrencia = DateUtil::getDataDeString($mes, 'd/m/Y')->format('Y/m/01 00:00:00');

            $resumos = $this->contabilService->getResumosByDtocorrenciaAndNrorg($dtocorrencia, $nrorg);

            foreach ($resumos as $resumo) {
                //$estrutura_legal = $this->empresaService->getEstruturaLegalSuperior($resumo['NRESTRUTURAM'], $dtocorrencia, $nrorg);
                $estrutura = $this->empresaService->getEstrutura($resumo['NRESTRUTURAMSUP'], $dtocorrencia, $nrorg);

                if (empty($estrutura)) {
                    continue;
                }

                if ($estrutura['CDCNPJESTRUT'] == $cnpj) {
                    $result[] = array(
                        'id' => $resumo['NRRESUMOCTBL'],
                        'num' => $resumo['NRLANCAMENTO'],
                        'tipo' => $resumo['IDTIPOLANC'],
                        'data' => $resumo['DTULTATU'],
                        'cod_respons' => $resumo['CDINTESTRUTEXP'],
                        'cod_unidade' => $resumo['CDINTESTRUTEXPFIL'],
                        'conta' => $resumo['CDCONTCTBL'],
                        'desc_conta' => $resumo['NMCONTCTBL'],
                        'historico' => $resumo['CDHISTPADR'],
                        'desc_hist' => $resumo['DSHISTPADR'],
                        'valor' => $resumo['VRLANCAMENTO'],
                    );
                }
            }

            if (empty($result))
                throw new \Exception("Não foram encontrados lançamentos contábeis para os parâmetros informados.");

            return $result;
        } catch (\Exception $e) {
            throw $e;
        }
    }

    public function findResumoDroByCnpjAndDataAndMovimento($filial, $data, $movimento)
    {
        try {
            $result = array();

            $nrorg = $this->environment->getNrOrg();

            $nrtipomoviment = $this->calculoService->getNumeroTipoMovimento($movimento);

            $resumos = $this->contabilService->getResumosDroByCnpjSupAndDtocorrenciaAndNrorg($nrorg, $filial, $data, $nrtipomoviment);

            foreach ($resumos as $resumo) {
                $result[] = array(
                    'data' => $resumo['DTOCORRENCIA'],
                    'cod_unidade' => $resumo['CDINTESTRUTEXPFIL'],
                    'cod_despesa' => $resumo['CDDESPADMFAT'],
                    'desc_despesa' => $resumo['DSDESPADMFAT'],
                    'valor' => $resumo['VRLANCAMENTO'],
                    'qtd_func' => $resumo['NRQTDEVINC'],
                );
            }

            if (empty($result))
                throw new \Exception("Não foram encontrados lançamentos de despesas para os parâmetros informados.");

            return $result;
        } catch (\Exception $e) {
            throw $e;
        }
    }

    public function retornaDadosPessoa($dtmescompetenc, $cdcnpjestrut, $cdfilial, $cdcentcust, $nrcpfpessoa, $dtreferencia, $cdintegocupa){
        try {
            $result = array();

            $nrorg = $this->environment->getNrOrg();
            $dtmescompetenc = !is_null($dtmescompetenc) ? DateUtil::getDataDeString($dtmescompetenc)->format('01/m/Y') : DateUtil::getDataAtual(true)->format('01/m/Y');
            $dtreferencia = !is_null($dtreferencia) ? DateUtil::getDataDeString($dtreferencia)->format('d/m/Y') : null;

            /* Retorna pessoas */
            $retorno = $this->pessoaService->getDadosPessoa(
                $nrorg, $dtmescompetenc, $cdcnpjestrut, $cdfilial, $cdcentcust, $nrcpfpessoa, $dtreferencia, $cdintegocupa
            );

            if (empty($retorno))
                throw new \Exception("Não foram encontradas pessoas para os parâmetros informados.");

            return $retorno;
        }
        catch (\Exception $e) {
            throw $e;
        }
    }
    
    public function setLancamentoLote($dslotelancevento, $dtcompetloteeve, $nrtipomoviment, $dsobslotelanc, $cdfilial, $cdcentrocusto, $nrcpfpessoa, 
                                      $nreventom, $vrreffolhalote, $vreventolote, $dtocorrenciaitem, $dsitemlotelanc)
    {
        try {
            $this->entityManager->beginTransaction();

            $lote = false;
            $item = false;

            $nrorg = $this->environment->getNrOrg();
            $nrorgpadrao = $this->environment->getNrOrgPadrao();
            $cdoperador = self::CDOPERADOR;
            $dtcompetloteeve = !is_null($dtcompetloteeve) ? DateUtil::getDataDeString($dtcompetloteeve)->format('01/m/Y') : DateUtil::getDataAtual(true)->format('01/m/Y');
            $dtocorrenciaitem = !is_null($dtocorrenciaitem) ? DateUtil::getDataDeString($dtocorrenciaitem)->format('d/m/Y') : null;
            
            // Busca o lote para a descrição informada
            $lotes = $this->eventoService->getLotelancevento($nrorg, $dslotelancevento, $dtcompetloteeve, $nrtipomoviment);
            
            // Se não existir o lote cria um novo
            if (empty($lotes)) {
                $nrlotelancevento = NovoCodigo::geraCodigo($nrorg, "FPA_LOTELANCEVENTO", 1, 20, true, false);
                $lote = $this->eventoService->addLotelancevento(
                    $nrorg, $nrlotelancevento, $dslotelancevento, $dtcompetloteeve, $nrtipomoviment, $dsobslotelanc, $cdoperador
                );
            } else {
                $lote = true;
                $nrlotelancevento = $lotes[0]['NRLOTELANCEVENTO'];
            }
            
            // Busca vinculo
            $vinculo = $this->pessoaService->getVinculoByCpfAndFilialandCentroCusto($nrorg, $dtcompetloteeve, $nrcpfpessoa, $cdfilial, $cdcentrocusto);

            if (empty($vinculo))
                throw new \Exception("Não foi encontrado vínculo para o CPF informado ($nrcpfpessoa).");
            
            // Busca evento
            $evento = $this->eventoService->getEvento($nrorg, $nrorgpadrao, $dtcompetloteeve, $nreventom);

            if (empty($evento))
                throw new \Exception("Não foi encontrado o evento informado ($nreventom).");
            
            if (is_array($vinculo[0]) && is_array($evento[0])) {

                $itemlotelanc = $this->eventoService->getItemlotelanc($nrorg, $nrlotelancevento, $vinculo[0]['NRVINCULOM'], $evento[0]['NREVENTOM'], $dtocorrenciaitem);

                if (!empty($itemlotelanc)) {
                    $vrreffolhalote += $itemlotelanc[0]['VRREFFOLHALOTE'];
                    $vreventolote += $itemlotelanc[0]['VREVENTOLOTE'];
                    $item = $this->eventoService->updateItemLotelanc(
                        $itemlotelanc[0]['NRORG'], $itemlotelanc[0]['NRITLOTELANC'], $itemlotelanc[0]['NRLOTELANCEVENTO'], $vrreffolhalote, $vreventolote, $cdoperador
                    );
                } else {
                    //insere o item de lote
                    $nritemlotelanc = NovoCodigo::geraCodigo($nrorg, "FPA_ITEMLOTELANC", 1, 20, true, false);
                    $item = $this->eventoService->addItemLotelanc(
                        $nrorg, $nrlotelancevento, $nritemlotelanc, $vinculo[0]['NRVINCULOM'], $evento[0]['NREVENTOM'], $vrreffolhalote, $vreventolote, $dtocorrenciaitem, $dsitemlotelanc, $cdoperador
                    );
                }
            }
            
            if ($lote && $item) {
                $this->entityManager->commit();
            } else {
                throw new \Exception("Erro inesperado ao incluir o lote e/ou seus itens");
            }

            return true;
        } catch (\Exception $e) {
            $this->entityManager->rollback();
            throw $e;
        }
    }
    
    public function setMovimentacaoGerencialVinculo($cdfilial, $cdcentrocusto, $nrcpfpessoa, $dtinimoviment, $dtfimmoviment)
    {
        try {
            $this->entityManager->beginTransaction();

            $nrorg = $this->environment->getNrOrg();
            $nrorgpadrao = $this->environment->getNrOrgPadrao();
            $cdoperador = self::CDOPERADOR;
            $dtinimoviment = !is_null($dtinimoviment) ? DateUtil::getDataDeString($dtinimoviment)->format('d/m/Y') : DateUtil::getDataAtual(true)->format('01/m/Y');
            $dtfimmoviment = !is_null($dtfimmoviment) ? DateUtil::getDataDeString($dtfimmoviment)->format('d/m/Y') : null;
            $dtmescompetencia = DateUtil::getDataDeString($dtinimoviment)->format('01/m/Y');
            
            // Busca vinculo
            $vinculo = $this->pessoaService->getVinculoByCpfAndFilial($nrorg, $dtmescompetencia, $nrcpfpessoa, $cdfilial);

            if (empty($vinculo))
                throw new \Exception("Não foi encontrado vínculo para o CPF e Filial informados ($nrcpfpessoa, $cdfilial).");
            
            // Busca estrutura de destino
            $estrutura = $this->empresaService->getEstruturaRelacionadaByFilialAndCentrocusto($nrorg, $dtmescompetencia, $vinculo[0]['NRESTRUTLEGAL'], $cdfilial, $cdcentrocusto);

            if (empty($estrutura))
                throw new \Exception("Não foi encontrada estrutura para o Centro de Custro informado ($cdcentrocusto).");

            if ($vinculo[0]['NRESTRUTGEREN'] == $estrutura[0]['NRESTRUTURAM'])
                throw new \Exception("O vínculo " . $vinculo[0]['NRVINCULOM'] . " já está alocado na estrutura " . $estrutura[0]['NRESTRUTURAM'] . ".");
            
            if (is_array($vinculo[0]) && is_array($estrutura[0])) {

                $movimentacaoGerencial = $this->pessoaService->getMovimentacaoGerencial($nrorg, $dtmescompetencia, $vinculo[0]['NRVINCULOM']);

                if (DateUtil::getDataDeString($movimentacaoGerencial[0]['DTINIMOVIMENT'], 'Y-m-d H:i:s', true)->format('d/m/Y') == $dtinimoviment)
                    throw new \Exception("Vínculo já possui movimentação gerencial na data informada.");

                if (!empty($movimentacaoGerencial)) {
                    // Atualiza a data final da movimentação gerencial atual com o dia anterior da nova movimentação
                    $dtfim = DateUtil::subtraiIntervalo(DateUtil::getDataDeString($dtinimoviment, DateUtil::FORMATO_BRASILEIRO, true), 1)->format(DateUtil::FORMATO_BRASILEIRO);
                    $this->pessoaService->updateDataFimMovimentacao(
                        $movimentacaoGerencial[0]['NRORG'], $movimentacaoGerencial[0]['NRMOVIMENTACAO'], $dtfim, $cdoperador
                    );
                }
                
                // Insere nova movimentação gerencial
                $nrmovimentacao = NovoCodigo::geraCodigo($nrorg, "GPE_MOVIMENTACAO", 1, 20, true, false);
                $this->pessoaService->addMovimentacao(
                    $nrorg, $nrmovimentacao, $vinculo[0]['NRVINCULOM'], 2, 8, $estrutura[0]['NRTIPOESTRUTURA'], $estrutura[0]['NRESTRUTURAM'], $dtinimoviment, $dtfimmoviment, 
                    null, 'HcmServices', $cdoperador
                );
                
                // Atualiza o histórico do vínculo
                // Caso não tenha histórico para a competência, cria um novo
                $this->pessoaService->atualizaHistoricoVinculo($nrorg, $vinculo[0]['NRVINCULOM'], $dtinimoviment, $cdoperador, true, true, null, $estrutura[0]['NRESTRUTURAM']);
                
                $this->entityManager->commit();
                return true;
            }
            
            throw new \Exception("Erro inesperado ao incluir a movimentação");
        } catch (\Exception $e) {
            $this->entityManager->rollback();
            throw $e;
        }
    }

    public function setEstrutura($nrtipoestrutura, $dtinivigencia, $nmfantasia, $nmrazsocestrut, $cdcnpjestrut, 
        $nrcependereco, $nmpais, $sgestado, $nmmunicipio, $nmbairroendereco, $cdlogradouro, $dsendereco, $nrimovelendereco, $dscompleendereco, $dsreferenciaende, 
        $idtipoempr, $idoptsimples, $idparticipat, $nrtipotribut, $nrclasstribut, $vraliqdeson, $nrcalendario, $cdcnae, $cdfpas, $cdterceiro, $vrfap, $cdceiestrut, 
        $nrcaged, $cdemprcaixa, $cdfederalestrut, $nrpropriempr, $nrparcrespinf, $nrparcentisin
    )
    {
        try {
            $this->entityManager->beginTransaction();

            $nrorg = $this->environment->getNrOrg();
            $nrorgpadrao = $this->environment->getNrOrgPadrao();
            $cdoperador = self::CDOPERADOR;
            $dtinivigencia = !is_null($dtinivigencia) ? DateUtil::getDataDeString($dtinivigencia)->format('d/m/Y') : DateUtil::getDataAtual(true)->format('01/m/Y');
            $dtmescompetenc = DateUtil::getDataDeString($dtinivigencia)->format('01/m/Y');

            // Verifica se já existe estrutura para o mesmo tipo de estrutura e mesmo CNPJ
            $estrutura = $this->empresaService->getEstruturaByCnpjAndTipoestrutura($nrorg, $dtmescompetenc, $cdcnpjestrut, $nrtipoestrutura);
            if (!empty($estrutura))
                throw new \Exception("Já existe estrutura para o mesmo tipo de estrutura e CNPJ");
            
            // Busca País
            $pais = $this->geralService->getPaisByNome($nmpais);
            if (empty($pais))
                throw new \Exception("Não foi encontrado país para o nome '$nmpais'");
            
            // Busca Município
            $municipio = $this->geralService->getMunicipioByPaisAndEstadoAndNome($pais['CDPAIS'], $sgestado, $nmmunicipio);
            if (empty($municipio))
                throw new \Exception("Não foi encontrado município para o nome '$nmmunicipio'");

            // Busca Bairro
            $bairro = $this->geralService->getBairroByPaisAndEstadoAndMunicipioAndNome($pais['CDPAIS'], $sgestado, $municipio['CDMUNICIPIO'], $nmbairroendereco);
            $nmbairro = empty($bairro) ? null : $bairro['NMBAIRRO'];

            // Busca o pessoa do proprietário da empresa
            if (!is_null($nrpropriempr)) {
                $propriempr = $this->pessoaService->getPessoaByCpfAndCompetenciaAndNrorg($nrpropriempr, $dtmescompetenc, $nrorg);
                if (empty($propriempr)) {
                    throw new \Exception("Não foi encontrada pessoa proprietária da empresa para o CPF '$nrpropriempr'");
                } else {
                    $nrpropriempr = $propriempr['NRPARCNEGOCIO'];
                }
            }

            // Busca o pessoa do responsável informal da empresa
            if (!is_null($nrparcrespinf)) {
                $parcrespinf = $this->pessoaService->getPessoaByCpfAndCompetenciaAndNrorg($nrparcrespinf, $dtmescompetenc, $nrorg);
                if (empty($parcrespinf)) {
                    throw new \Exception("Não foi encontrada pessoa responsável informal da empresa para o CPF '$nrparcrespinf'");
                } else {
                    $nrparcrespinf = $parcrespinf['NRPARCNEGOCIO'];
                }
            }

            // Busca o pessoa do responsável legal da empresa
            if (!is_null($nrparcentisin)) {
                $parcentisin = $this->pessoaService->getPessoaByCpfAndCompetenciaAndNrorg($nrparcentisin, $dtmescompetenc, $nrorg);
                if (empty($parcentisin)) {
                    throw new \Exception("Não foi encontrada pessoa responsável legal da empresa para o CPF '$nrparcentisin'");
                } else {
                    $nrparcentisin = $parcentisin['NRPARCNEGOCIO'];
                }
            }

            // Insere o parceiro da estrutura
            $nrparcnegocio = NovoCodigo::geraCodigo($nrorg, "PARCNEGOCIO", 1, 20, true, false);
            $this->geralService->addParceiroNegocio($nrorg, $nrparcnegocio, $nmrazsocestrut, $nmrazsocestrut, 'N', 'S', 'N',  $dtinivigencia, $cdcnpjestrut, 'ESTRUTURA', 'CNPJ', $cdoperador);

            // Insere o endereço da estrutura
            $nrenderecoparc = NovoCodigo::geraCodigo($nrorg, "ENDERECOPARC", 1, 20, true, false);
            $this->geralService->addEnderecoParceiroNegocio($nrorg, $nrenderecoparc, $nrparcnegocio, 'PRINCIPAL', $pais['CDPAIS'], $sgestado, $municipio['CDMUNICIPIO'], 
                                                            $bairro['CDBAIRRO'], $cdlogradouro, $nmbairroendereco, $dscompleendereco, $nrcependereco, $dsreferenciaende,
                                                            $dsendereco, $nrimovelendereco, $cdoperador);

            // Insere o mestre da estrutura
            $nrestruturam = NovoCodigo::geraCodigo($nrorg, "ESTRUTURAM", 1, 20, true, false);
            $this->empresaService->addEstruturaMestre($nrorg, $nrestruturam, $nrtipoestrutura, $dtinivigencia, $nmfantasia, $nrparcnegocio, null, null, 
                                                      null, null, null, null, null, $cdoperador);

            // Insere o histórico da estrutura
            $nrestruturah = NovoCodigo::geraCodigo($nrorg, "ESTRUTURAH", 1, 20, true, false);
            $this->empresaService->addEstruturaHistorico($nrorg, $nrestruturah, $nrestruturam, $dtmescompetenc, $cdcnpjestrut, $cdceiestrut, $nmrazsocestrut, 
                    null, $cdcnae, $nrcaged, $cdemprcaixa, $idtipoempr, $idoptsimples, $idparticipat, $nrparcnegocio, $nrpropriempr, $nrparcrespinf, $nrparcentisin, 
                    $cdfpas, $nmfantasia, $nmrazsocestrut, null, null, $cdterceiro, $nrclasstribut, $vrfap, $nrtipotribut, 
                    $vraliqdeson, $cdfederalestrut, null, null, null, null, $cdoperador);

            $this->entityManager->commit();
            return true;
        } catch (\Exception $e) {
            $this->entityManager->rollback();
            throw $e;
        }
    }
    
    public function retornaMarcacoes($nrcpfpessoa, $dtiniper, $dtfimper, $cdfilial, $cdcentrocusto){
        try {
            
            if(!$cdfilial && !$cdcentrocusto && !$nrcpfpessoa){
                throw new \Exception("Parãmetros de estrutura ou cpf não informados.");
            }
            
            $result = array();
            
            $nrorg = $this->environment->getNrOrg();
            $dtmescompetenc = DateUtil::getDataAtual(true)->format('01/m/Y');

            /* Retorna Marcações de ponto */
            $retorno = $this->pessoaService->getMarcacao($nrorg, $dtmescompetenc, $nrcpfpessoa, $dtiniper, $dtfimper, $cdfilial, $cdcentrocusto);

            if (empty($retorno))
                throw new \Exception("Não foram encontradas marcações para os parâmetros informados.");

            return $retorno;
        }
        catch (\Exception $e) {
            throw $e;
        }
    }
    
}